@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2000-2002  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2003-2006  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2003-2007  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -33,6 +33,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <getopt.h>
+#include <sys/poll.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
@@ -75,20 +76,16 @@ enum {
 	WRITE,
 	RECEIVE,
 	SEND,
+	SERVER,
 	PPPDUMP,
 	AUDIO
 };
 
 /* Default options */
-static int  device;
 static int  snap_len = SNAP_LEN;
-static int  defpsm = 0;
-static int  defcompid = DEFAULT_COMPID;
 static int  mode = PARSE;
 static int  permcheck = 1;
 static int  noappend = 0;
-static long flags;
-static long filter;
 static char *dump_file = NULL;
 static char *pppdump_file = NULL;
 static char *audio_file = NULL;
@@ -127,9 +124,16 @@ static uint8_t btsnoop_id[] = { 0x62, 0x74, 0x73, 0x6e, 0x6f, 0x6f, 0x70, 0x00 }
 static uint32_t btsnoop_version = 0;
 static uint32_t btsnoop_type = 0;
 
+struct pktlog_hdr {
+	uint32_t	len;
+	uint64_t	ts;
+	uint8_t		type;
+} __attribute__ ((packed));
+#define PKTLOG_HDR_SIZE (sizeof(struct pktlog_hdr))
+
 static inline int read_n(int fd, char *buf, int len)
 {
-	register int t = 0, w;
+	int t = 0, w;
 
 	while (len > 0) {
 		if ((w = read(fd, buf, len)) < 0) {
@@ -146,7 +150,7 @@ static inline int read_n(int fd, char *buf, int len)
 
 static inline int write_n(int fd, char *buf, int len)
 {
-	register int t = 0, w;
+	int t = 0, w;
 
 	while (len > 0) {
 		if ((w = write(fd, buf, len)) < 0) {
@@ -161,7 +165,7 @@ static inline int write_n(int fd, char *buf, int len)
 	return t;
 }
 
-static void process_frames(int dev, int sock, int fd, unsigned long flags)
+static int process_frames(int dev, int sock, int fd, unsigned long flags)
 {
 	struct cmsghdr *cmsg;
 	struct msghdr msg;
@@ -169,9 +173,17 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 	struct hcidump_hdr *dh;
 	struct btsnoop_pkt *dp;
 	struct frame frm;
+	struct pollfd fds[2];
+	int nfds = 0;
 	char *buf, *ctrl;
 	int len, hdr_size = HCIDUMP_HDR_SIZE;
 
+	if (sock < 0)
+		return -1;
+
+	if (mode == SERVER)
+		flags |= DUMP_BTSNOOP;
+
 	if (snap_len < SNAP_LEN)
 		snap_len = SNAP_LEN;
 
@@ -181,7 +193,7 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 	buf = malloc(snap_len + hdr_size);
 	if (!buf) {
 		perror("Can't allocate data buffer");
-		exit(1);
+		return -1;
 	}
 
 	dh = (void *) buf;
@@ -190,8 +202,9 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 
 	ctrl = malloc(100);
 	if (!ctrl) {
+		free(buf);
 		perror("Can't allocate control buffer");
-		exit(1);
+		return -1;
 	}
 
 	if (dev == HCI_DEV_NONE)
@@ -199,11 +212,72 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 	else
 		printf("device: hci%d ", dev);
 
-	printf("snap_len: %d filter: 0x%lx\n", snap_len, filter);
+	printf("snap_len: %d filter: 0x%lx\n", snap_len, parser.filter);
 
 	memset(&msg, 0, sizeof(msg));
 
+	if (mode == SERVER) {
+		struct btsnoop_hdr *hdr = (void *) buf;
+
+		btsnoop_version = 1;
+		btsnoop_type = 1002;
+
+		memcpy(hdr->id, btsnoop_id, sizeof(btsnoop_id));
+		hdr->version = htonl(btsnoop_version);
+		hdr->type = htonl(btsnoop_type);
+
+		printf("btsnoop version: %d datalink type: %d\n",
+						btsnoop_version, btsnoop_type);
+
+		len = write(fd, buf, BTSNOOP_HDR_SIZE);
+		if (len < 0) {
+			perror("Can't create dump header");
+			return -1;
+		}
+
+		if (len != BTSNOOP_HDR_SIZE) {
+			fprintf(stderr, "Header size mismatch\n");
+			return -1;
+		}
+
+		fds[nfds].fd = fd;
+		fds[nfds].events = POLLIN;
+		fds[nfds].revents = 0;
+		nfds++;
+	}
+
+	fds[nfds].fd = sock;
+	fds[nfds].events = POLLIN;
+	fds[nfds].revents = 0;
+	nfds++;
+
 	while (1) {
+		int i, n = poll(fds, nfds, -1);
+		if (n <= 0)
+			continue;
+
+		for (i = 0; i < nfds; i++) {
+			if (fds[i].revents & (POLLHUP | POLLERR | POLLNVAL)) {
+				if (fds[i].fd == sock)
+					printf("device: disconnected\n");
+				else
+					printf("client: disconnect\n");
+				return 0;
+			}
+		}
+
+		if (mode == SERVER) {
+			len = recv(fd, buf, snap_len, MSG_DONTWAIT);
+			if (len == 0) {
+				printf("client: disconnect\n");
+				return 0;
+			}
+			if (len < 0 && errno != EAGAIN && errno != EINTR) {
+				perror("Connection read failure");
+				return -1;
+			}
+		}
+
 		iv.iov_base = frm.data;
 		iv.iov_len  = snap_len;
 
@@ -212,10 +286,12 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 		msg.msg_control = ctrl;
 		msg.msg_controllen = 100;
 
-		len = recvmsg(sock, &msg, 0);
+		len = recvmsg(sock, &msg, MSG_DONTWAIT);
 		if (len < 0) {
+			if (errno == EAGAIN || errno == EINTR)
+				continue;
 			perror("Receive failed");
-			exit(1);
+			return -1;
 		}
 
 		/* Process control message */
@@ -244,6 +320,7 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 		switch (mode) {
 		case WRITE:
 		case SEND:
+		case SERVER:
 			/* Save or send dump */
 			if (flags & DUMP_BTSNOOP) {
 				uint64_t ts;
@@ -266,7 +343,7 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 
 			if (write_n(fd, buf, frm.data_len + hdr_size) < 0) {
 				perror("Write error");
-				exit(1);
+				return -1;
 			}
 			break;
 
@@ -276,12 +353,15 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 			break;
 		}
 	}
+
+	return 0;
 }
 
 static void read_dump(int fd)
 {
 	struct hcidump_hdr dh;
 	struct btsnoop_pkt dp;
+	struct pktlog_hdr ph;
 	struct frame frm;
 	uint8_t pkt_type;
 	int err;
@@ -293,7 +373,9 @@ static void read_dump(int fd)
 	}
 
 	while (1) {
-		if (parser.flags & DUMP_BTSNOOP)
+		if (parser.flags & DUMP_PKTLOG)
+			err = read_n(fd, (void *) &ph, PKTLOG_HDR_SIZE);
+		else if (parser.flags & DUMP_BTSNOOP)
 			err = read_n(fd, (void *) &dp, BTSNOOP_PKT_SIZE);
 		else
 			err = read_n(fd, (void *) &dh, HCIDUMP_HDR_SIZE);
@@ -303,7 +385,32 @@ static void read_dump(int fd)
 		if (!err)
 			return;
 
-		if (parser.flags & DUMP_BTSNOOP) {
+		if (parser.flags & DUMP_PKTLOG) {
+			switch (ph.type) {
+			case 0x00:
+				((uint8_t *) frm.data)[0] = HCI_COMMAND_PKT;
+				frm.in = 0;
+				break;
+			case 0x01:
+				((uint8_t *) frm.data)[0] = HCI_EVENT_PKT;
+				frm.in = 1;
+				break;
+			case 0x02:
+				((uint8_t *) frm.data)[0] = HCI_ACLDATA_PKT;
+				frm.in = 0;
+				break;
+			case 0x03:
+				((uint8_t *) frm.data)[0] = HCI_ACLDATA_PKT;
+				frm.in = 1;
+				break;
+			default:
+				lseek(fd, ntohl(ph.len) - 9, SEEK_CUR);
+				continue;
+			}
+
+			frm.data_len = ntohl(ph.len) - 8;
+			err = read_n(fd, frm.data + 1, frm.data_len - 1);
+		} else if (parser.flags & DUMP_BTSNOOP) {
 			switch (btsnoop_type) {
 			case 1001:
 				if (ntohl(dp.flags) & 0x02) {
@@ -338,12 +445,17 @@ static void read_dump(int fd)
 		frm.ptr = frm.data;
 		frm.len = frm.data_len;
 
-		if (parser.flags & DUMP_BTSNOOP) {
+		if (parser.flags & DUMP_PKTLOG) {
+			uint64_t ts;
+			ts = ntoh64(ph.ts);
+			frm.ts.tv_sec = ts >> 32;
+			frm.ts.tv_usec = ts & 0xffffffff;
+		} else if (parser.flags & DUMP_BTSNOOP) {
 			uint64_t ts;
 			frm.in = ntohl(dp.flags) & 0x01;
 			ts = ntoh64(dp.ts) - 0x00E03AB44A676000ll;
 			frm.ts.tv_sec = (ts / 1000000ll) + 946684800ll;
-			frm.ts.tv_usec = ts % 1000000ll; 
+			frm.ts.tv_usec = ts % 1000000ll;
 		} else {
 			frm.in = dh.in;
 			frm.ts.tv_sec  = btohl(dh.ts_sec);
@@ -360,7 +472,8 @@ failed:
 
 static int open_file(char *file, int mode, unsigned long flags)
 {
-	struct btsnoop_hdr hdr;
+	unsigned char buf[BTSNOOP_HDR_SIZE];
+	struct btsnoop_hdr *hdr = (struct btsnoop_hdr *) buf;
 	int fd, len, open_flags;
 
 	if (mode == WRITE || mode == PPPDUMP || mode == AUDIO) {
@@ -378,17 +491,17 @@ static int open_file(char *file, int mode, unsigned long flags)
 	}
 
 	if (mode == READ) {
-		len = read(fd, &hdr, BTSNOOP_HDR_SIZE);
+		len = read(fd, buf, BTSNOOP_HDR_SIZE);
 		if (len != BTSNOOP_HDR_SIZE) {
 			lseek(fd, 0, SEEK_SET);
 			return fd;
 		}
 
-		if (!memcmp(hdr.id, btsnoop_id, sizeof(btsnoop_id))) {
+		if (!memcmp(hdr->id, btsnoop_id, sizeof(btsnoop_id))) {
 			parser.flags |= DUMP_BTSNOOP;
 
-			btsnoop_version = ntohl(hdr.version);
-			btsnoop_type = ntohl(hdr.type);
+			btsnoop_version = ntohl(hdr->version);
+			btsnoop_type = ntohl(hdr->type);
 
 			printf("btsnoop version: %d datalink type: %d\n",
 						btsnoop_version, btsnoop_type);
@@ -403,6 +516,11 @@ static int open_file(char *file, int mode, unsigned long flags)
 				exit(1);
 			}
 		} else {
+			if (buf[0] == 0x00 && buf[1] == 0x00) {
+				parser.flags |= DUMP_PKTLOG;
+				printf("packet logger data format\n");
+			}
+
 			parser.flags &= ~DUMP_BTSNOOP;
 			lseek(fd, 0, SEEK_SET);
 			return fd;
@@ -412,14 +530,14 @@ static int open_file(char *file, int mode, unsigned long flags)
 			btsnoop_version = 1;
 			btsnoop_type = 1002;
 
-			memcpy(hdr.id, btsnoop_id, sizeof(btsnoop_id));
-			hdr.version = htonl(btsnoop_version);
-			hdr.type = htonl(btsnoop_type);
+			memcpy(hdr->id, btsnoop_id, sizeof(btsnoop_id));
+			hdr->version = htonl(btsnoop_version);
+			hdr->type = htonl(btsnoop_type);
 
 			printf("btsnoop version: %d datalink type: %d\n",
 						btsnoop_version, btsnoop_type);
 
-			len = write(fd, &hdr, BTSNOOP_HDR_SIZE);
+			len = write(fd, buf, BTSNOOP_HDR_SIZE);
 			if (len < 0) {
 				perror("Can't create dump header");
 				exit(1);
@@ -446,19 +564,19 @@ static int open_socket(int dev, unsigned long flags)
 		dd = hci_open_dev(dev);
 		if (dd < 0) {
 			perror("Can't open device");
-			exit(1);
+			return -1;
 		}
 
 		if (hci_devinfo(dev, &di) < 0) {
 			perror("Can't get device info");
-			exit(1);
+			return -1;
 		}
 
 		opt = hci_test_bit(HCI_RAW, &di.flags);
 		if (ioctl(dd, HCISETRAW, opt) < 0) {
 			if (errno == EACCES) {
 				perror("Can't access device");
-				exit(1);
+				return -1;
 			}
 		}
 
@@ -469,19 +587,19 @@ static int open_socket(int dev, unsigned long flags)
 	sk = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI);
 	if (sk < 0) {
 		perror("Can't create raw socket");
-		exit(1);
+		return -1;
 	}
 
 	opt = 1;
 	if (setsockopt(sk, SOL_HCI, HCI_DATA_DIR, &opt, sizeof(opt)) < 0) {
 		perror("Can't enable data direction info");
-		exit(1);
+		return -1;
 	}
 
 	opt = 1;
 	if (setsockopt(sk, SOL_HCI, HCI_TIME_STAMP, &opt, sizeof(opt)) < 0) {
 		perror("Can't enable time stamp");
-		exit(1);
+		return -1;
 	}
 
 	/* Setup filter */
@@ -490,7 +608,7 @@ static int open_socket(int dev, unsigned long flags)
 	hci_filter_all_events(&flt);
 	if (setsockopt(sk, SOL_HCI, HCI_FILTER, &flt, sizeof(flt)) < 0) {
 		perror("Can't set filter");
-		exit(1);
+		return -1;
 	}
 
 	/* Bind socket to the HCI device */
@@ -499,7 +617,7 @@ static int open_socket(int dev, unsigned long flags)
 	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
 		printf("Can't attach to device hci%d. %s(%d)\n", 
 					dev, strerror(errno), errno);
-		exit(1);
+		return -1;
 	}
 
 	return sk;
@@ -510,12 +628,12 @@ static int open_connection(char *addr, char *port)
 	struct sockaddr_storage ss;
 	struct addrinfo hints, *res0, *res;
 	int sk = -1, opt = 1;
-	
+
 	memset(&hints, 0, sizeof(hints));
 	hints.ai_family = af;
 	hints.ai_socktype = SOCK_STREAM;
 	hints.ai_protocol = IPPROTO_TCP;
-	
+
 	if (getaddrinfo(addr, port, &hints, &res0))
 		if(getaddrinfo(NULL, port, &hints, &res0)) {
 			perror("getaddrinfo");
@@ -534,7 +652,7 @@ static int open_connection(char *addr, char *port)
 		}
 
 		setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
-		
+
 		memcpy(&ss, res->ai_addr, res->ai_addrlen);
 
 		switch(ss.ss_family) {
@@ -543,7 +661,8 @@ static int open_connection(char *addr, char *port)
 			((struct sockaddr_in *) &ss)->sin_port = 0;
 			break;
 		case AF_INET6:
-			memcpy(&((struct sockaddr_in6 *) &ss)->sin6_addr, &in6addr_any, sizeof(in6addr_any));
+			memcpy(&((struct sockaddr_in6 *) &ss)->sin6_addr,
+						&in6addr_any, sizeof(in6addr_any));
 			((struct sockaddr_in6 *) &ss)->sin6_port = 0;
 			break;
 		}
@@ -568,87 +687,197 @@ static int open_connection(char *addr, char *port)
 	return sk;
 }
 
+static int create_datagram(unsigned short port)
+{
+	struct sockaddr_in addr;
+	int sk, opt = 1;
+
+	sk = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
+	if (sk < 0)
+		return -1;
+
+	if (setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) < 0) {
+		close(sk);
+		return -1;
+	}
+
+	memset(&addr, 0, sizeof(addr));
+	addr.sin_family = AF_INET;
+	addr.sin_port = htons(port);
+	addr.sin_addr.s_addr = htonl(INADDR_BROADCAST);
+
+	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
+		close(sk);
+		return -1;
+	}
+
+	return sk;
+}
+
+static unsigned char ping_data[] = { 'p', 'i', 'n', 'g' };
+static unsigned char pong_data[] = { 'p', 'o', 'n', 'g' };
+
+static void handle_datagram(int sk)
+{
+	struct sockaddr_in addr;
+	socklen_t addr_len = sizeof(addr);
+	unsigned char buf[64];
+	ssize_t len;
+
+	len = recvfrom(sk, buf, sizeof(buf), MSG_DONTWAIT,
+				(struct sockaddr *) &addr, &addr_len);
+
+	if (len != sizeof(ping_data))
+		return;
+
+	if (memcmp(buf, ping_data, sizeof(ping_data)) != 0)
+		return;
+
+	len = sendto(sk, pong_data, sizeof(pong_data), 0,
+				(struct sockaddr *) &addr, sizeof(addr));
+}
+
 static int wait_connection(char *addr, char *port)
 {
 	char hname[100], hport[10];
-	struct sockaddr_storage ss;
-	struct addrinfo hints, *res0, *res;
-	socklen_t len;
-	int sk = -1, nsk, opt = 1;
+	struct addrinfo *ai, *runp;
+	struct addrinfo hints;
+	struct pollfd fds[3];
+	int err, opt, datagram, nfds = 0;
 
-	memset(&hints, 0, sizeof(hints));
-	hints.ai_family = af;
+	memset(&hints, 0, sizeof (hints));
+	hints.ai_flags = AI_PASSIVE | AI_ADDRCONFIG;
 	hints.ai_socktype = SOCK_STREAM;
 	hints.ai_protocol = IPPROTO_TCP;
 
-	if (getaddrinfo(addr, port, &hints, &res0))
-		if (getaddrinfo(NULL, port, &hints, &res0)) {
-			perror("getaddrinfo");
-			exit(1);
-		}
- 	
-	for (res = res0; res; res = res->ai_next) {
-		sk = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
-		if (sk < 0) {
-			if (res->ai_next)
-				continue;
+	err = getaddrinfo(dump_addr, dump_port, &hints, &ai);
+	if (err < 0) {
+		printf("Can't get address info: %s\n", gai_strerror(err));
+		return -1;
+	}
 
+	runp = ai;
+
+	datagram = create_datagram(atoi(dump_port));
+	if (datagram < 0) {
+		printf("server: no discover protocol\n");
+	} else {
+		fds[nfds].fd = datagram;
+		fds[nfds].events = POLLIN;
+		nfds++;
+	}
+
+	while (runp != NULL && nfds < sizeof(fds) / sizeof(fds[0])) {
+		fds[nfds].fd = socket(runp->ai_family, runp->ai_socktype,
+							runp->ai_protocol);
+		if (fds[nfds].fd < 0) {
 			perror("Can't create socket");
- 			freeaddrinfo(res0);
-			exit(1);
+			return -1;
 		}
 
-		setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
+		fds[nfds].events = POLLIN;
 
-		if (bind(sk, res->ai_addr, res->ai_addrlen) < 0) {
-			if(res->ai_next) {
-				close(sk);
+		opt = 1;
+		setsockopt(fds[nfds].fd, SOL_SOCKET, SO_REUSEADDR,
+							&opt, sizeof(opt));
+
+		opt = 0;
+		setsockopt(fds[nfds].fd, SOL_SOCKET, SO_KEEPALIVE,
+							&opt, sizeof(opt));
+
+		if (bind(fds[nfds].fd, runp->ai_addr, runp->ai_addrlen) < 0) {
+			if (errno != EADDRINUSE) {
+				perror("Can't bind socket");
+				return -1;
+			}
+
+			close(fds[nfds].fd);
+		} else {
+			if (listen(fds[nfds].fd, SOMAXCONN) < 0) {
+				perror("Can't listen on socket");
+				return -1;
+			}
+
+			getnameinfo(runp->ai_addr, runp->ai_addrlen,
+							hname, sizeof(hname),
+							hport, sizeof(hport),
+							NI_NUMERICSERV);
+
+			printf("server: %s:%s snap_len: %d filter: 0x%lx\n",
+					hname, hport, snap_len, parser.filter);
+
+			nfds++;
+		}
+
+		runp = runp->ai_next;
+	}
+
+	freeaddrinfo(ai);
+
+	while (1) {
+		int i, n = poll(fds, nfds, -1);
+		if (n <= 0)
+			continue;
+
+		for (i = 0; i < nfds; i++) {
+			struct sockaddr_storage rem;
+			socklen_t remlen = sizeof(rem);
+			int sk;
+
+			if (!(fds[i].revents & POLLIN))
+				continue;
+
+			if (fds[i].fd == datagram) {
+				handle_datagram(datagram);
 				continue;
 			}
 
-			perror("Can't bind socket");
-			close(sk);
-			freeaddrinfo(res0);
-			exit(1);
- 		}
-
-		getnameinfo(res->ai_addr, res->ai_addrlen, hname, sizeof(hname),
-					hport, sizeof(hport), NI_NUMERICSERV);
-		printf("device: %s:%s snap_len: %d filter: 0x%lx\n", 
-					hname, hport, snap_len, filter);
-		if (listen(sk, 1) < 0) {
-			if (res->ai_next) {
-				close(sk);
+			sk = accept(fds[i].fd, (struct sockaddr *) &rem, &remlen);
+			if (sk < 0)
 				continue;
-			}
 
-			perror("Can't listen on socket");
-			close(sk);
-			freeaddrinfo(res0);
-			exit(1);
+			getnameinfo((struct sockaddr *) &rem, remlen,
+							hname, sizeof(hname),
+							hport, sizeof(hport),
+							NI_NUMERICSERV);
+
+			printf("client: %s:%s snap_len: %d filter: 0x%lx\n",
+					hname, hport, snap_len, parser.filter);
+
+			for (n = 0; n < nfds; n++)
+				close(fds[n].fd);
+
+			return sk;
 		}
 	}
 
-	freeaddrinfo(res0);
+	return -1;
+}
 
-	len = sizeof(ss);
-	nsk = accept(sk, (struct sockaddr *) &ss, &len);
-	if (nsk < 0) {
-		perror("Can't accept new socket");
+static int run_server(int dev, char *addr, char *port, unsigned long flags)
+{
+	while (1) {
+		int dd, sk;
+
+		sk = wait_connection(addr, port);
+		if (sk < 0)
+			continue;
+
+		//fcntl(sk, F_SETFL, O_NONBLOCK);
+
+		dd = open_socket(dev, flags);
+		if (dd < 0) {
+			close(sk);
+			continue;
+		}
+
+		process_frames(dev, dd, sk, flags);
+
+		close(dd);
 		close(sk);
-		freeaddrinfo(res0);
-		exit(1);
 	}
 
-	getnameinfo((struct sockaddr *) &ss, sizeof(ss),
-					hname, sizeof(hname), NULL, 0, 0);
-
-	printf("device: %s snap_len: %d filter: 0x%lx\n",
-						hname, snap_len, filter);
-
-	close(sk);
-
-	return nsk;
+	return 0;
 }
 
 static struct {
@@ -675,8 +904,9 @@ static struct {
 	{ 0 }
 };
 
-static void parse_filter(int argc, char **argv)
+static unsigned long parse_filter(int argc, char **argv)
 {
+	unsigned long filter = 0;
 	int i,n;
 
 	for (i = 0; i < argc; i++) {
@@ -687,6 +917,8 @@ static void parse_filter(int argc, char **argv)
 			}
 		}
 	}
+
+	return filter;
 }
 
 static void usage(void)
@@ -701,6 +933,7 @@ static void usage(void)
 	"  -r, --read-dump=file       Read dump from a file\n"
 	"  -s, --send-dump=host       Send dump to a host\n"
 	"  -n, --recv-dump=host       Receive dump on a host\n"
+	"  -d, --wait-dump=host       Wait on a host and send\n"
 	"  -t, --ts                   Display time stamps\n"
 	"  -a, --ascii                Dump data in ascii\n"
 	"  -x, --hex                  Dump data in hex\n"
@@ -732,6 +965,7 @@ static struct option main_options[] = {
 	{ "read-dump",		1, 0, 'r' },
 	{ "send-dump",		1, 0, 's' },
 	{ "recv-dump",		1, 0, 'n' },
+	{ "wait-dump",		1, 0, 'd' },
 	{ "timestamp",		0, 0, 't' },
 	{ "ascii",		0, 0, 'a' },
 	{ "hex",		0, 0, 'x' },
@@ -756,11 +990,16 @@ static struct option main_options[] = {
 
 int main(int argc, char *argv[])
 {
+	unsigned long flags = 0;
+	unsigned long filter = 0;
+	int device = 0;
+	int defpsm = 0;
+	int defcompid = DEFAULT_COMPID;
 	int opt, pppdump_fd = -1, audio_fd = -1;
 
 	printf("HCI sniffer - Bluetooth packet analyzer ver %s\n", VERSION);
 
-	while ((opt=getopt_long(argc, argv, "i:l:p:m:w:r:s:n:taxXRC:H:O:P:D:A:BVYZN46h", main_options, NULL)) != -1) {
+	while ((opt=getopt_long(argc, argv, "i:l:p:m:w:r:s:n:d:taxXRC:H:O:P:D:A:BVYZN46h", main_options, NULL)) != -1) {
 		switch(opt) {
 		case 'i':
 			if (strcasecmp(optarg, "none") && strcasecmp(optarg, "system"))
@@ -801,6 +1040,11 @@ int main(int argc, char *argv[])
 			dump_addr = optarg;
 			break;
 
+		case 'd':
+			mode = SERVER;
+			dump_addr = optarg;
+			break;
+
 		case 't': 
 			flags |= DUMP_TSTAMP;
 			break;
@@ -885,7 +1129,7 @@ int main(int argc, char *argv[])
 	optind = 0;
 
 	if (argc > 0)
-		parse_filter(argc, argv);
+		filter = parse_filter(argc, argv);
 
 	/* Default settings */
 	if (!filter)
@@ -922,6 +1166,11 @@ int main(int argc, char *argv[])
 		process_frames(device, open_socket(device, flags),
 				open_connection(dump_addr, dump_port), flags);
 		break;
+
+	case SERVER:
+		init_parser(flags, filter, defpsm, defcompid, pppdump_fd, audio_fd);
+		run_server(device, dump_addr, dump_port, flags);
+		break;
 	}
 
 	return 0;
